#include <libsoup/soup.h>
#include <libsoup/soup-xmlrpc-message.h>
#include "soup-md5-utils.h"
#include <glib/gthread.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>
#include <libosso.h>
#include <gconf/gconf-client.h>
#if defined(TABLET_BROWSER)
# include <tablet-browser-interface.h>
#elif defined(OSSO_BROWSER)
# include <osso-browser-interface.h>
#endif
#ifdef HILDON
# if HILDON == 1
#  include <hildon/hildon-window.h>
#  include <hildon/hildon-defines.h>
#  include <hildon/hildon-banner.h>
# else
#  include <hildon-widgets/hildon-window.h>
#  include <hildon-widgets/hildon-defines.h>
#  include <hildon-widgets/hildon-banner.h>
# endif
#else
# define HILDON_MARGIN_DEFAULT 5
#endif

#define API_KEY "73fc51d360159c5bb4a8f147270cf0c9"
#define API_SECRET "2537d871c9392998"

struct PixbufWriteData {
	GtkWidget *progbar;
	GString *string;
	FILE *file;
	gsize data_written;
};

static gboolean string_append_data(const gchar *bytes,
				   gsize len,
				   GError **error,
				   gpointer user_data)
{
	struct PixbufWriteData *data = user_data;

	(void)error;

	if (data->file) {
		fwrite(bytes, 1, len, data->file);
	} else {
		g_string_append_len(data->string, bytes, len);
	}

	data->data_written += len;

	if (len > 0x800) {
		gtk_progress_bar_pulse(GTK_PROGRESS_BAR(data->progbar));
		while (gtk_events_pending())
			gtk_main_iteration();
	}

	return TRUE;
}

static gchar *pixbuf_png_pack(GdkPixbuf *buf,
			      const gchar *file,
			      gsize *len,
			      GtkWidget *parent)
{
	struct PixbufWriteData data = { NULL, NULL, NULL, 0 };
	GtkWidget *banner;
	GError *error = NULL;

	data.progbar = gtk_progress_bar_new();

	if (file) {
		data.file = fopen(file, "w");
	} else {
		data.string = g_string_new("");
	}

	gtk_progress_bar_pulse(GTK_PROGRESS_BAR(data.progbar));
	banner = hildon_banner_show_progress(parent,
					     GTK_PROGRESS_BAR(data.progbar),
					     "Compressing...");

	while (gtk_events_pending())
		gtk_main_iteration();

	gdk_pixbuf_save_to_callback(buf,
				    string_append_data,
				    &data,
				    "png",
				    &error,
				    NULL);

	if (error != NULL) {
		g_error_free(error);
	}

	if (len) {
		*len = data.data_written;
	}

	gtk_widget_destroy(banner);

	return data.string ? g_string_free(data.string, FALSE) : NULL;
}


struct form_multipart_builder
{
	char *boundary;
	GString *content;
	SoupMD5Context *ctx;
};

static void _build_list(gpointer key, gpointer value, gpointer user_data)
{
	GList **list = user_data;
	(void)value;
	*list = g_list_append(*list, key);
}

static GList *_g_hash_table_get_keys(GHashTable *table)
{
	GList *list = NULL;
	g_hash_table_foreach(table, _build_list, &list);
	return list;
}

static SoupXmlrpcMessage *flickr_soup_call(const gchar *method,
			      const gchar *secret,
			      const gchar *first_arg,
			      ...)
{
	SoupXmlrpcMessage *msg = soup_xmlrpc_message_new("http://api.flickr.com/services/xmlrpc/");
	va_list args;
	GHashTable *hash_table = g_hash_table_new(g_str_hash, g_str_equal);
	GList *keys;
	SoupMD5Context ctx;
	va_start(args, first_arg);
	while (first_arg) {
		const gchar *value = va_arg(args, const gchar *);
		g_hash_table_replace(hash_table,
				     (gpointer)first_arg,
				     (gpointer)value);
		first_arg = va_arg(args, const gchar *);
	}
	va_end(args);

	keys = _g_hash_table_get_keys(hash_table);
	keys = g_list_sort(keys, (GCompareFunc)g_utf8_collate);

	soup_xmlrpc_message_start_call(msg, method);
	soup_xmlrpc_message_start_param(msg);
	soup_xmlrpc_message_start_struct(msg);
	if (secret) {
		soup_md5_init(&ctx);
		soup_md5_update(&ctx, secret, strlen(secret));
	}
	while (keys) {
		const gchar *key = keys->data;
		const gchar *value = g_hash_table_lookup(hash_table, key);
		if (secret) {
			soup_md5_update(&ctx, key, strlen(key));
			soup_md5_update(&ctx, value, strlen(value));
		}
		soup_xmlrpc_message_start_member(msg, key);
		soup_xmlrpc_message_write_string(msg, value);
		soup_xmlrpc_message_end_member(msg);
		keys = g_list_delete_link(keys, keys);
	}
	if (secret) {
		char md5sum[33];
		soup_xmlrpc_message_start_member(msg, "api_sig");
		soup_md5_final_hex(&ctx, md5sum);
		soup_xmlrpc_message_write_string(msg, md5sum);
		soup_xmlrpc_message_end_member(msg);
	}
	soup_xmlrpc_message_end_struct(msg);
	soup_xmlrpc_message_end_param(msg);
	soup_xmlrpc_message_end_call(msg);
	soup_xmlrpc_message_persist(msg);

	return msg;
}

static struct form_multipart_builder *form_multipart_builder_new(const gchar *shared)
{
	struct form_multipart_builder *builder =
		g_new0(struct form_multipart_builder, 1);
	srand(time(NULL));
	builder->boundary =
		g_strdup_printf("-----------------------------%06x",
				rand());
	builder->content = g_string_new("");
	if (shared) {
		builder->ctx = g_new0(SoupMD5Context, 1);
		soup_md5_init(builder->ctx);
		soup_md5_update(builder->ctx, shared, strlen(shared));
	}

	return builder;
}

static void form_multipart_builder_append_string(struct form_multipart_builder *builder,
						 const gchar *name,
						 const gchar *value)
{
	g_string_append_printf(builder->content,
			       "--%s\r\n"
			       "Content-Disposition: form-data; "
			       "name=\"%s\"\r\n\r\n"
			       "%s\r\n",
			       builder->boundary,
			       name,
			       value);
	if (builder->ctx) {
		soup_md5_update(builder->ctx, name, strlen(name));
		soup_md5_update(builder->ctx, value, strlen(value));
	}
}

static void form_multipart_builder_append_file_data(struct form_multipart_builder *builder,
						    const gchar *name,
						    const gchar *mime_type,
						    const gchar *file,
						    const gchar *data,
						    gssize data_len)
{
	g_string_append_printf(builder->content,
			       "--%s\r\n"
			       "Content-Disposition: form-data; "
			       "name=\"%s\"; "
			       "filename=\"%s\"\r\n"
			       "Content-type: %s\r\n\r\n",
			       builder->boundary,
			       name,
			       file,
			       mime_type);
	g_string_append_len(builder->content,
			    data,
			    (gssize)data_len);
	g_string_append(builder->content, "\r\n");
}

/*static void form_multipart_builder_append_file(struct form_multipart_builder *builder,
					       const gchar *name,
					       const gchar *mime_type,
					       const gchar *file)
{
	gchar *file_contents;
	gsize file_len;

	g_file_get_contents(file,
			    &file_contents,
			    &file_len,
			    NULL);

	if (file_contents)
	{
		form_multipart_builder_append_file_data(builder,
							name,
							mime_type,
							file,
							file_contents,
							(gssize)file_len);
		g_free(file_contents);
	}
} */

static void form_multipart_builder_to_message(struct form_multipart_builder *builder,
					      SoupMessage *message)
{
	gulong len;
	gchar *content_type = g_strdup_printf("multipart/form-data; "
					      "boundary=%s",
					      builder->boundary);
	if (builder->ctx) {
		char md5sum[33];
		soup_md5_final_hex(builder->ctx, md5sum);
		g_string_append_printf(builder->content,
				       "--%s\r\n"
				       "Content-Disposition: form-data; "
				       "name=api_sig\r\n\r\n"
				       "%s\r\n",
				       builder->boundary,
				       md5sum);
		g_free(builder->ctx);
		builder->ctx = NULL;
	}

	g_string_append_printf(builder->content,
			       "--%s--\r\n", 
			       builder->boundary);
	len = builder->content->len;
	soup_message_set_request(message,
				 content_type,
				 SOUP_BUFFER_SYSTEM_OWNED,
				 g_string_free(builder->content, FALSE),
				 len);
	g_free(content_type);
	g_free(builder->boundary);
	g_free(builder);
}

enum Responses {
	SHOT_RESPONSE_CANCEL = 1,
	SHOT_RESPONSE_OK,
};

static gboolean flickr_save(GdkPixbuf *buf, GtkWindow *parent,
			    osso_context_t *osso)
{
	SoupMessage *message;
	SoupXmlrpcMessage *msg;
	SoupXmlrpcResponse *response;
	SoupXmlrpcValue *value;
	SoupSession *session;
	SoupMD5Context ctx;
	GtkWidget *dialog;
	char md5sum[33];
	gchar *frob = NULL;
	gchar *buf_data;
	gsize buf_size;
	char buffer[1024];
	struct form_multipart_builder *builder;
	time_t now = time(NULL);
	struct tm times;

	localtime_r(&now, &times);
	strftime(buffer, 1023, "shot-%Y-%m-%d-%H-%M-%S.png", &times);

	session = soup_session_sync_new();
	msg = flickr_soup_call("flickr.auth.getFrob",
			       API_SECRET,
			       "api_key",
			       API_KEY,
			       NULL);
	soup_session_send_message(session, (SoupMessage *)msg);
	response = soup_xmlrpc_message_parse_response(msg);

	value = soup_xmlrpc_response_get_value(response);
	if (soup_xmlrpc_value_get_type(value) ==
	    SOUP_XMLRPC_VALUE_TYPE_STRING) {
		gchar *uri;
		char *code;
		soup_xmlrpc_value_get_string(value, &code);
		g_strstrip(code);
		code[strlen(code) - 7] = '\0';
		soup_md5_init(&ctx);
		soup_md5_update(&ctx, API_SECRET, 16);
		soup_md5_update(&ctx, "api_key", 7);
		soup_md5_update(&ctx, API_KEY, 32);
		soup_md5_update(&ctx, "frob", 4);
		soup_md5_update(&ctx, code + 6, strlen(code + 6));
		soup_md5_update(&ctx, "perms", 5);
		soup_md5_update(&ctx, "write", 5);
		soup_md5_final_hex(&ctx, md5sum);
		uri = g_strdup_printf("http://flickr.com/services/auth/?api_key=73fc51d360159c5bb4a8f147270cf0c9&perms=write&frob=%s&api_sig=%s\n", code + 6, md5sum);
		osso_rpc_run_with_defaults(osso, "osso_browser",
					   OSSO_BROWSER_OPEN_NEW_WINDOW_REQ,
					   NULL,
					   DBUS_TYPE_STRING, uri,
					   DBUS_TYPE_INVALID);
		frob = g_strdup(code + 6);
		free(code);
	}

	dialog = gtk_dialog_new_with_buttons("Waiting",
					     parent,
					     GTK_DIALOG_NO_SEPARATOR |
					     GTK_DIALOG_MODAL,
					     "Done",
					     GTK_RESPONSE_OK,
					     NULL);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
			  gtk_label_new("Please authenticate"));
	gtk_widget_show_all(GTK_WIDGET(dialog));
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	if (!frob) exit(0);
	msg = flickr_soup_call("flickr.auth.getToken",
			       API_SECRET,
			       "api_key", API_KEY,
			       "frob", frob,
			       NULL);

	soup_session_send_message(session, (SoupMessage *)msg);
	response = soup_xmlrpc_message_parse_response(msg);
	value = soup_xmlrpc_response_get_value(response);
	if (soup_xmlrpc_value_get_type(value) ==
	    SOUP_XMLRPC_VALUE_TYPE_STRING) {
		char *code;
		char *token;
		soup_xmlrpc_value_get_string(value, &code);
		if ((token = strstr(code, "</token>")))
			*token = '\0';
		token = strstr(code, "<token>");
		if (token) {
			g_free(frob);
			frob = g_strdup(token + 7);
		}
		free(code);
	}

	message = soup_message_new("POST",
				   "http://api.flickr.com/services/upload/");
	builder = form_multipart_builder_new(API_SECRET);
	form_multipart_builder_append_string(builder, "api_key", API_KEY);
	form_multipart_builder_append_string(builder, "auth_token", frob);
	form_multipart_builder_append_string(builder, "content_type", "2");
	buf_data = pixbuf_png_pack(buf,
				   NULL,
				   &buf_size,
				   GTK_WIDGET(parent));
	form_multipart_builder_append_file_data(builder, "photo", "image/png", buffer, buf_data, (gssize)buf_size);
	form_multipart_builder_to_message(builder, message);

	soup_session_send_message(session, message);
	return TRUE;
}

/*
gboolean picasa_save(GdkPixbuf *buf, GtkWindow *parent)
{
	return TRUE;
}
*/

static gboolean disk_save(GdkPixbuf *buf, GtkWindow *parent,
			  osso_context_t *osso)
{
	gchar buffer[1024];
	GtkWidget *chooser = gtk_file_chooser_dialog_new("Save as...",
							 parent,
							 GTK_FILE_CHOOSER_ACTION_SAVE,
							 "Save", 
							 SHOT_RESPONSE_OK,
							 "Cancel",
							 SHOT_RESPONSE_CANCEL,
							 NULL);
	time_t now = time(NULL);
	struct tm times;
	gchar *last_path;
	gchar *save;
	GConfClient *gc;

	(void)osso;

	gc = gconf_client_get_default();

	last_path = gconf_client_get_string(gc,
					    "/apps/mh-shot-tool/disk/path",
					    NULL);

	localtime_r(&now, &times);
	strftime(buffer, 1023, "shot-%Y-%m-%d-%H-%M-%S.png", &times);
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(chooser), buffer);
	if (last_path)
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(chooser),
						    last_path);
	gtk_widget_show_all(chooser);

	switch (gtk_dialog_run(GTK_DIALOG(chooser))) {
	case SHOT_RESPONSE_OK:
		save = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(chooser));
		pixbuf_png_pack(buf, save, NULL, GTK_WIDGET(parent));
		gconf_client_set_string(gc,
					"/apps/mh-shot-tool/disk/path",
					gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(chooser)),
					NULL);
		gtk_widget_destroy(chooser);
		g_object_unref(gc);
		return TRUE;
		break;
	case SHOT_RESPONSE_CANCEL:
	default:
		g_object_unref(gc);
		return FALSE;
		break;
	}
}

struct handle_pixbuf {
	const char *name;
	gboolean (*save)(GdkPixbuf *buf, GtkWindow *parent,
			 osso_context_t *osso);
} handlers[] = {
	{ "Flickr this", flickr_save },
/*	{ "Picasa this", picasa_save }, */
	{ "Save as...", disk_save },
	{ NULL, NULL }
};

static void _my_quit(GObject *object, gint *ret)
{
	gpointer data = g_object_get_data(object, "response");
	*ret = GPOINTER_TO_INT(data);
	gtk_main_quit();
}

struct GrabbedData {
	GdkPixbuf *buf;
	GtkWidget *image;
	GtkWidget *window;
	GtkWidget *spinner;
};

static gboolean do_grab(struct GrabbedData *grab)
{
	GdkPixbuf *thumb;
	grab->buf = gdk_pixbuf_get_from_drawable(NULL,
						 GDK_DRAWABLE(gdk_get_default_root_window()),
						 gdk_colormap_get_system(),
						 0, 0,
						 0, 0,
						 gdk_screen_width(),
						 gdk_screen_height());
	thumb = gdk_pixbuf_scale_simple(grab->buf,
					gdk_screen_width() / 4,
					gdk_screen_height() / 4,
					GDK_INTERP_BILINEAR);

	gtk_image_set_from_pixbuf(GTK_IMAGE(grab->image),
			     thumb);
	g_object_unref(thumb);

	gtk_widget_show_all(grab->window);

	return FALSE;
}

static void _regrab(GtkWidget *button, struct GrabbedData *grab)
{
	(void)button;

	g_object_unref(grab->buf);
	g_timeout_add(gtk_spin_button_get_value(GTK_SPIN_BUTTON(grab->spinner)) * 1000,
		      (GSourceFunc)do_grab,
		      grab);
	gtk_widget_hide(grab->window);
}

int main(int argc, char **argv)
{
	GConfClient *gc;
	struct GrabbedData grab;
	GtkWidget *combo;
	GtkWidget *vbox;
	GtkWidget *hbox;
	osso_context_t *osso;
	guint i;
	GtkWidget *button;
	gint ret = GTK_RESPONSE_NONE;
	gchar *last_handler;

	g_set_application_name("MH Shot Tool");
	g_thread_init(NULL);
	gtk_init(&argc, &argv);

	osso = osso_initialize("org.MaemoHackers.MhShotTool", "0.0.1",
			       FALSE, NULL);
#ifdef HILDON
	grab.window = hildon_window_new();
#else
	grab.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#endif
	g_signal_connect(grab.window, "delete-event",
			 G_CALLBACK(gtk_main_quit), NULL);
	gtk_window_set_title(GTK_WINDOW(grab.window), "Take screenshot");
	vbox = gtk_vbox_new(FALSE, 0);

	grab.image = gtk_image_new();
	gtk_container_add(GTK_CONTAINER(vbox), grab.image);

	hbox = gtk_hbox_new(FALSE, HILDON_MARGIN_DEFAULT);
	button = gtk_button_new_with_label("Re-grab");
	g_signal_connect(button, "clicked", G_CALLBACK(_regrab), &grab);
	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("with"), FALSE, FALSE, 0);
	grab.spinner = gtk_spin_button_new_with_range(0.0, 3600.0, 0.5);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(grab.spinner), 5.0);
	gtk_box_pack_start(GTK_BOX(hbox), grab.spinner, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("second interval"), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
	
	hbox = gtk_hbutton_box_new();
	button = gtk_button_new_with_label("Ok");
	g_object_set_data(G_OBJECT(button), "response",
			  GINT_TO_POINTER(SHOT_RESPONSE_OK));
	g_signal_connect(button, "clicked", G_CALLBACK(_my_quit), &ret);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	button = gtk_button_new_with_label("Cancel");
	g_object_set_data(G_OBJECT(button), "response",
			  GINT_TO_POINTER(SHOT_RESPONSE_CANCEL));
	g_signal_connect(button, "clicked", G_CALLBACK(_my_quit), &ret);
	gtk_container_add(GTK_CONTAINER(hbox), button);
					     
	gc = gconf_client_get_default();
	last_handler = gconf_client_get_string(gc,
					       "/apps/mh-shot-tool/handler",
					       NULL);
	
	combo = gtk_combo_box_new_text();
	for (i = 0; handlers[i].name; i++) {
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo),
					  handlers[i].name);
		if (last_handler && !strcmp(handlers[i].name, last_handler)) {
			gtk_combo_box_set_active(GTK_COMBO_BOX(combo), i);
		}
	}
	if (gtk_combo_box_get_active(GTK_COMBO_BOX(combo)) == -1)
		gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);

	gtk_box_pack_start(GTK_BOX(vbox), combo, FALSE, TRUE, 0);
	gtk_box_pack_end(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(grab.window), vbox);

	(void)do_grab(&grab);

	gtk_main();

	switch (ret)
	{
	case SHOT_RESPONSE_OK:
		i = gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
		handlers[i].save(grab.buf, GTK_WINDOW(grab.window), osso);
		gconf_client_set_string(gc,
					"/apps/mh-shot-tool/handler",
					handlers[i].name,
					NULL);
		break;
	case SHOT_RESPONSE_CANCEL:
	default:
		break;
	}
	g_object_unref(grab.buf);
	g_object_unref(gc);
	g_free(last_handler);
	osso_deinitialize(osso);

	return 0;
}
